@@ -9,7 +9,9 @@ import warnings
 
 import numpy as np
 import torch
+from torch.optim.lr_scheduler import ReduceLROnPlateau
 
+from pytorch_widedeep.metrics import MultipleMetrics
 from pytorch_widedeep.wdtypes import *  # noqa: F403
 
 
@@ -17,6 +19,14 @@ def _get_current_time():
     return datetime.datetime.now().strftime("%B %d, %Y - %I:%M%p")
 
 
+def _is_metric(monitor: str):
+    # We assume no one will use f3 or more
+    if any([s in monitor for s in ["acc", "prec", "rec", "fscore", "f1", "f2"]]):
+        return True
+    else:
+        return False
+
+
 class CallbackContainer(object):
     """
     Container holding a list of callbacks.
@@ -52,10 +62,12 @@ class CallbackContainer(object):
         for callback in self.callbacks:
             callback.on_epoch_begin(epoch, logs)
 
-    def on_epoch_end(self, epoch: int, logs: Optional[Dict] = None):
+    def on_epoch_end(
+        self, epoch: int, logs: Optional[Dict] = None, metric: Optional[float] = None
+    ):
         logs = logs or {}
         for callback in self.callbacks:
-            callback.on_epoch_end(epoch, logs)
+            callback.on_epoch_end(epoch, logs, metric)
 
     def on_batch_begin(self, batch: int, logs: Optional[Dict] = None):
         logs = logs or {}
@@ -81,6 +93,12 @@ class CallbackContainer(object):
         for callback in self.callbacks:
             callback.on_train_end(logs)
 
+    def on_eval_begin(self, logs: Optional[Dict] = None):
+        # at the moment only used to reset metrics before eval
+        logs = logs or {}
+        for callback in self.callbacks:
+            callback.on_eval_begin(logs)
+
 
 class Callback(object):
     """
@@ -102,7 +120,9 @@ class Callback(object):
     def on_epoch_begin(self, epoch: int, logs: Optional[Dict] = None):
         pass
 
-    def on_epoch_end(self, epoch: int, logs: Optional[Dict] = None):
+    def on_epoch_end(
+        self, epoch: int, logs: Optional[Dict] = None, metric: Optional[float] = None
+    ):
         pass
 
     def on_batch_begin(self, batch: int, logs: Optional[Dict] = None):
@@ -117,48 +137,129 @@ class Callback(object):
     def on_train_end(self, logs: Optional[Dict] = None):
         pass
 
+    def on_eval_begin(self, logs: Optional[Dict] = None):
+        # at the moment only used to reset metrics before eval
+        pass
+
 
 class History(Callback):
     r"""Callback that records metrics to a ``history`` attribute.
 
     This callback runs by default within :obj:`Trainer`, therefore, should not
-    be passed to the ``Trainer``. Is included here just for completion.
+    be passed to the :obj:`Trainer`. Is included here just for completion.
     """
 
     def on_train_begin(self, logs: Optional[Dict] = None):
         self.trainer.history = {}
 
-    def on_epoch_end(self, epoch: int, logs: Optional[Dict] = None):
+    def on_epoch_end(
+        self, epoch: int, logs: Optional[Dict] = None, metric: Optional[float] = None
+    ):
         logs = logs or {}
         for k, v in logs.items():
+            if isinstance(v, np.ndarray):
+                v = v.tolist()
             self.trainer.history.setdefault(k, []).append(v)
 
 
+class LRShedulerCallback(Callback):
+    r"""Callback for the learning rate schedulers to take a step
+
+    This callback runs by default within :obj:`Trainer`, therefore, should not
+    be passed to the :obj:`Trainer`. Is included here just for completion.
+    """
+
+    def on_batch_end(self, batch: int, logs: Optional[Dict] = None):
+        if self.trainer.lr_scheduler is not None:
+            if self._multiple_scheduler():
+                for (
+                    model_name,
+                    scheduler,
+                ) in self.trainer.lr_scheduler._schedulers.items():
+                    if self._is_cyclic(model_name):
+                        scheduler.step()
+            elif self.trainer.cyclic_lr:
+                self.trainer.lr_scheduler.step()
+
+    def on_epoch_end(
+        self, epoch: int, logs: Optional[Dict] = None, metric: Optional[float] = None
+    ):
+        if self.trainer.lr_scheduler is not None:
+            if self._multiple_scheduler():
+                for (
+                    model_name,
+                    scheduler,
+                ) in self.trainer.lr_scheduler._schedulers.items():
+                    if not self._is_cyclic(model_name):
+                        if isinstance(scheduler, ReduceLROnPlateau):
+                            scheduler.step(metric)
+                        else:
+                            scheduler.step()
+            elif not self.trainer.cyclic_lr:
+                if isinstance(self.trainer.lr_scheduler, ReduceLROnPlateau):
+                    self.trainer.lr_scheduler.step(metric)
+                else:
+                    self.trainer.lr_scheduler.step()
+
+    def _multiple_scheduler(self):
+        return self.trainer.lr_scheduler.__class__.__name__ == "MultipleLRScheduler"
+
+    def _is_cyclic(self, model_name: str):
+        return (
+            self._has_scheduler(model_name)
+            and "cycl"
+            in self.trainer.lr_scheduler._schedulers[
+                model_name
+            ].__class__.__name__.lower()
+        )
+
+    def _has_scheduler(self, model_name: str):
+        return model_name in self.trainer.lr_scheduler._schedulers
+
+
+class MetricCallback(Callback):
+    r"""Callback that resets the metrics (if any metric is used)
+
+    This callback runs by default within :obj:`Trainer`, therefore, should not
+    be passed to the :obj:`Trainer`. Is included here just for completion.
+    """
+
+    def __init__(self, container: MultipleMetrics):
+        self.container = container
+
+    def on_epoch_begin(self, epoch: int, logs: Optional[Dict] = None):
+        self.container.reset()
+
+    def on_eval_begin(self, logs: Optional[Dict] = None):
+        self.container.reset()
+
+
 class LRHistory(Callback):
+    r"""Saves the learning rates during training to a ``lr_history`` attribute.
+
+    Callbacks are passed as input parameters to the :obj:`Trainer` class. See
+    :class:`pytorch_widedeep.trainer.Trainer`
+
+    Parameters
+    ----------
+    n_epochs: int
+        number of epochs durint training
+
+    Examples
+    --------
+    >>> from pytorch_widedeep.callbacks import LRHistory
+    >>> from pytorch_widedeep.models import TabMlp, Wide, WideDeep
+    >>> from pytorch_widedeep.training import Trainer
+    >>>
+    >>> embed_input = [(u, i, j) for u, i, j in zip(["a", "b", "c"][:4], [4] * 3, [8] * 3)]
+    >>> column_idx = {k: v for v, k in enumerate(["a", "b", "c"])}
+    >>> wide = Wide(10, 1)
+    >>> deep = TabMlp(mlp_hidden_dims=[8, 4], column_idx=column_idx, embed_input=embed_input)
+    >>> model = WideDeep(wide, deep)
+    >>> trainer = Trainer(model, objective="regression", callbacks=[LRHistory(n_epochs=10)])
+    """
+
     def __init__(self, n_epochs: int):
-        r"""Saves the learning rates during training to a ``lr_history`` attribute.
-
-        Callbacks are passed as input parameters to the ``Trainer`` class. See
-        :class:`pytorch_widedeep.trainer.Trainer`
-
-        Parameters
-        ----------
-        n_epochs: int
-            number of epochs durint training
-
-        Examples
-        --------
-        >>> from pytorch_widedeep.callbacks import LRHistory
-        >>> from pytorch_widedeep.models import TabMlp, Wide, WideDeep
-        >>> from pytorch_widedeep.training import Trainer
-        >>>
-        >>> embed_input = [(u, i, j) for u, i, j in zip(["a", "b", "c"][:4], [4] * 3, [8] * 3)]
-        >>> column_idx = {k: v for v, k in enumerate(["a", "b", "c"])}
-        >>> wide = Wide(10, 1)
-        >>> deep = TabMlp(mlp_hidden_dims=[8, 4], column_idx=column_idx, embed_input=embed_input)
-        >>> model = WideDeep(wide, deep)
-        >>> trainer = Trainer(model, objective="regression", callbacks=[LRHistory(n_epochs=10)])
-        """
         super(LRHistory, self).__init__()
         self.n_epochs = n_epochs
 
@@ -177,7 +278,9 @@ class LRHistory(Callback):
             elif self.trainer.cyclic_lr:
                 self._save_group_lr(self.trainer.optimizer)
 
-    def on_epoch_end(self, epoch: int, logs: Optional[Dict] = None):
+    def on_epoch_end(
+        self, epoch: int, logs: Optional[Dict] = None, metric: Optional[float] = None
+    ):
         if epoch != (self.n_epochs - 1) and self.trainer.lr_scheduler is not None:
             if self._multiple_scheduler():
                 self._save_group_lr_mulitple_scheduler(step_location="on_epoch_end")
@@ -220,6 +323,62 @@ class LRHistory(Callback):
 
 
 class ModelCheckpoint(Callback):
+    r"""Saves the model after every epoch.
+
+    This class is almost identical to the corresponding keras class.
+    Therefore, **credit** to the Keras Team.
+
+    Callbacks are passed as input parameters to the :obj:`Trainer` class. See
+    :class:`pytorch_widedeep.trainer.Trainer`
+
+    Parameters
+    ----------
+    filepath: str
+        Full path to save the output weights. It must contain only the root of
+        the filenames. Epoch number and ``.pt`` extension (for pytorch) will
+        be added. e.g. ``filepath="path/to/output_weights/weights_out"`` And
+        the saved files in that directory will be named: ``weights_out_1.pt,
+        weights_out_2.pt, ...``
+    monitor: str, default="loss"
+        quantity to monitor. Typically 'val_loss' or metric name (e.g. 'val_acc')
+    verbose:int, default=0,
+        verbosity mode
+    save_best_only: bool, default=False,
+        the latest best model according to the quantity monitored will not be
+        overwritten.
+    mode: str, default="auto",
+        If ``save_best_only=True``, the decision to overwrite the current save
+        file is made based on either the maximization or the minimization of
+        the monitored quantity. For `'acc'`, this should be `'max'`, for
+        `'loss'` this should be `'min'`, etc. In `'auto'` mode, the
+        direction is automatically inferred from the name of the monitored
+        quantity.
+    period: int, default=1,
+        Interval (number of epochs) between checkpoints.
+    max_save: int, default=-1
+        Maximum number of outputs to save. If -1 will save all outputs
+
+    Attributes
+    ----------
+    best: float
+        best metric
+    best_epoch: int
+        best epoch
+
+    Examples
+    --------
+    >>> from pytorch_widedeep.callbacks import ModelCheckpoint
+    >>> from pytorch_widedeep.models import TabMlp, Wide, WideDeep
+    >>> from pytorch_widedeep.training import Trainer
+    >>>
+    >>> embed_input = [(u, i, j) for u, i, j in zip(["a", "b", "c"][:4], [4] * 3, [8] * 3)]
+    >>> column_idx = {k: v for v, k in enumerate(["a", "b", "c"])}
+    >>> wide = Wide(10, 1)
+    >>> deep = TabMlp(mlp_hidden_dims=[8, 4], column_idx=column_idx, embed_input=embed_input)
+    >>> model = WideDeep(wide, deep)
+    >>> trainer = Trainer(model, objective="regression", callbacks=[ModelCheckpoint(filepath='checkpoints/weights_out')])
+    """
+
     def __init__(
         self,
         filepath: str,
@@ -230,56 +389,6 @@ class ModelCheckpoint(Callback):
         period: int = 1,
         max_save: int = -1,
     ):
-        r"""Saves the model after every epoch.
-
-        This class is almost identical to the corresponding keras class.
-        Therefore, **credit** to the Keras Team.
-
-        Callbacks are passed as input parameters to the ``Trainer`` class. See
-        :class:`pytorch_widedeep.trainer.Trainer`
-
-        Parameters
-        ----------
-        filepath: str
-            Full path to save the output weights. It must contain only the root of
-            the filenames. Epoch number and ``.pt`` extension (for pytorch) will
-            be added. e.g. ``filepath="path/to/output_weights/weights_out"`` And
-            the saved files in that directory will be named: ``weights_out_1.pt,
-            weights_out_2.pt, ...``
-        monitor: str, default="val_loss"
-            quantity to monitor. :obj:`ModelCheckpoint` will infer if this is a
-            loss (i.e. contains the str `'loss'`) or a metric (i.e. contains the
-            str `'acc'` or starts with `'fmeasure'`).
-        verbose:int, default=0,
-            verbosity mode
-        save_best_only: bool, default=False,
-            the latest best model according to the quantity monitored will not be
-            overwritten.
-        mode: str, default="auto",
-            If ``save_best_only=True``, the decision to overwrite the current save
-            file is made based on either the maximization or the minimization of
-            the monitored quantity. For `'val_acc'`, this should be `'max'`, for
-            `'val_loss'` this should be `'min'`, etc. In `'auto'` mode, the
-            direction is automatically inferred from the name of the monitored
-            quantity.
-        period: int, default=1,
-            Interval (number of epochs) between checkpoints.
-        max_save: int, default=-1
-            Maximum number of outputs to save. If -1 will save all outputs
-
-        Examples
-        --------
-        >>> from pytorch_widedeep.callbacks import ModelCheckpoint
-        >>> from pytorch_widedeep.models import TabMlp, Wide, WideDeep
-        >>> from pytorch_widedeep.training import Trainer
-        >>>
-        >>> embed_input = [(u, i, j) for u, i, j in zip(["a", "b", "c"][:4], [4] * 3, [8] * 3)]
-        >>> column_idx = {k: v for v, k in enumerate(["a", "b", "c"])}
-        >>> wide = Wide(10, 1)
-        >>> deep = TabMlp(mlp_hidden_dims=[8, 4], column_idx=column_idx, embed_input=embed_input)
-        >>> model = WideDeep(wide, deep)
-        >>> trainer = Trainer(model, objective="regression", callbacks=[ModelCheckpoint(filepath='checkpoints/weights_out')])
-        """
         super(ModelCheckpoint, self).__init__()
 
         self.filepath = filepath
@@ -319,14 +428,16 @@ class ModelCheckpoint(Callback):
             self.monitor_op = np.greater
             self.best = -np.Inf
         else:
-            if "acc" in self.monitor or self.monitor.startswith("fmeasure"):
+            if _is_metric(self.monitor):
                 self.monitor_op = np.greater
                 self.best = -np.Inf
             else:
                 self.monitor_op = np.less
                 self.best = np.Inf
 
-    def on_epoch_end(self, epoch: int, logs: Optional[Dict] = None):  # noqa: C901
+    def on_epoch_end(  # noqa: C901
+        self, epoch: int, logs: Optional[Dict] = None, metric: Optional[float] = None
+    ):
         logs = logs or {}
         self.epochs_since_last_save += 1
         if self.epochs_since_last_save >= self.period:
@@ -355,6 +466,7 @@ class ModelCheckpoint(Callback):
                                 )
                             )
                         self.best = current
+                        self.best_epoch = epoch
                         torch.save(self.model.state_dict(), filepath)
                         if self.max_save > 0:
                             if len(self.old_files) == self.max_save:
@@ -383,8 +495,72 @@ class ModelCheckpoint(Callback):
                         self.old_files = self.old_files[1:]
                     self.old_files.append(filepath)
 
+    def __getstate__(self):
+        d = self.__dict__
+        self_dict = {k: d[k] for k in d if k not in ["trainer", "model"]}
+        return self_dict
+
+    def __setstate__(self, state):
+        self.__dict__ = state
+
 
 class EarlyStopping(Callback):
+    r"""Stop training when a monitored quantity has stopped improving.
+
+    This class is almost identical to the corresponding keras class.
+    Therefore, **credit** to the Keras Team.
+
+    Callbacks are passed as input parameters to the :obj:`Trainer` class. See
+    :class:`pytorch_widedeep.trainer.Trainer`
+
+    Parameters
+    -----------
+    monitor: str, default='val_loss'.
+        Quantity to monitor. Typically 'val_loss' or metric name (e.g. 'val_acc')
+    min_delta: float, default=0.
+        minimum change in the monitored quantity to qualify as an
+        improvement, i.e. an absolute change of less than min_delta, will
+        count as no improvement.
+    patience: int, default=10.
+        Number of epochs that produced the monitored quantity with no
+        improvement after which training will be stopped.
+    verbose: int.
+        verbosity mode.
+    mode: str, default='auto'
+        one of {'`auto`', '`min`', '`max`'}. In `'min'` mode, training will
+        stop when the quantity monitored has stopped decreasing; in `'max'`
+        mode it will stop when the quantity monitored has stopped increasing;
+        in `'auto'` mode, the direction is automatically inferred from the
+        name of the monitored quantity.
+    baseline: float, Optional. default=None.
+        Baseline value for the monitored quantity to reach. Training will
+        stop if the model does not show improvement over the baseline.
+    restore_best_weights: bool, default=None
+        Whether to restore model weights from the epoch with the best
+        value of the monitored quantity. If ``False``, the model weights
+        obtained at the last step of training are used.
+
+    Attributes
+    ----------
+    best: float
+        best metric
+    stopped_epoch: int
+        epoch when the training stopped
+
+    Examples
+    --------
+    >>> from pytorch_widedeep.callbacks import EarlyStopping
+    >>> from pytorch_widedeep.models import TabMlp, Wide, WideDeep
+    >>> from pytorch_widedeep.training import Trainer
+    >>>
+    >>> embed_input = [(u, i, j) for u, i, j in zip(["a", "b", "c"][:4], [4] * 3, [8] * 3)]
+    >>> column_idx = {k: v for v, k in enumerate(["a", "b", "c"])}
+    >>> wide = Wide(10, 1)
+    >>> deep = TabMlp(mlp_hidden_dims=[8, 4], column_idx=column_idx, embed_input=embed_input)
+    >>> model = WideDeep(wide, deep)
+    >>> trainer = Trainer(model, objective="regression", callbacks=[EarlyStopping(patience=10)])
+    """
+
     def __init__(
         self,
         monitor: str = "val_loss",
@@ -395,54 +571,6 @@ class EarlyStopping(Callback):
         baseline: Optional[float] = None,
         restore_best_weights: bool = False,
     ):
-        r"""Stop training when a monitored quantity has stopped improving.
-
-        This class is almost identical to the corresponding keras class.
-        Therefore, **credit** to the Keras Team.
-
-        Callbacks are passed as input parameters to the ``Trainer`` class. See
-        :class:`pytorch_widedeep.trainer.Trainer`
-
-        Parameters
-        -----------
-        monitor: str, default='val_loss'.
-            Quantity to be monitored.
-        min_delta: float, default=0.
-            minimum change in the monitored quantity to qualify as an
-            improvement, i.e. an absolute change of less than min_delta, will
-            count as no improvement.
-        patience: int, default=10.
-            Number of epochs that produced the monitored quantity with no
-            improvement after which training will be stopped.
-        verbose: int.
-            verbosity mode.
-        mode: str, default='auto'
-            one of {'`auto`', '`min`', '`max`'}. In `'min'` mode, training will
-            stop when the quantity monitored has stopped decreasing; in `'max'`
-            mode it will stop when the quantity monitored has stopped increasing;
-            in `'auto'` mode, the direction is automatically inferred from the
-            name of the monitored quantity.
-        baseline: float, Optional. default=None.
-            Baseline value for the monitored quantity to reach. Training will
-            stop if the model does not show improvement over the baseline.
-        restore_best_weights: bool, default=None
-            Whether to restore model weights from the epoch with the best
-            value of the monitored quantity. If ``False``, the model weights
-            obtained at the last step of training are used.
-
-        Examples
-        --------
-        >>> from pytorch_widedeep.callbacks import EarlyStopping
-        >>> from pytorch_widedeep.models import TabMlp, Wide, WideDeep
-        >>> from pytorch_widedeep.training import Trainer
-        >>>
-        >>> embed_input = [(u, i, j) for u, i, j in zip(["a", "b", "c"][:4], [4] * 3, [8] * 3)]
-        >>> column_idx = {k: v for v, k in enumerate(["a", "b", "c"])}
-        >>> wide = Wide(10, 1)
-        >>> deep = TabMlp(mlp_hidden_dims=[8, 4], column_idx=column_idx, embed_input=embed_input)
-        >>> model = WideDeep(wide, deep)
-        >>> trainer = Trainer(model, objective="regression", callbacks=[EarlyStopping(patience=10)])
-        """
         super(EarlyStopping, self).__init__()
 
         self.monitor = monitor
@@ -470,7 +598,7 @@ class EarlyStopping(Callback):
         elif self.mode == "max":
             self.monitor_op = np.greater
         else:
-            if "acc" in self.monitor:
+            if _is_metric(self.monitor):
                 self.monitor_op = np.greater
             else:
                 self.monitor_op = np.less
@@ -489,7 +617,9 @@ class EarlyStopping(Callback):
         else:
             self.best = np.Inf if self.monitor_op == np.less else -np.Inf
 
-    def on_epoch_end(self, epoch: int, logs: Optional[Dict] = None):
+    def on_epoch_end(
+        self, epoch: int, logs: Optional[Dict] = None, metric: Optional[float] = None
+    ):
         current = self.get_monitor_value(logs)
         if current is None:
             return
@@ -506,9 +636,7 @@ class EarlyStopping(Callback):
                 self.trainer.early_stop = True
                 if self.restore_best_weights:
                     if self.verbose > 0:
-                        print(
-                            "Restoring model weights from the end of " "the best epoch"
-                        )
+                        print("Restoring model weights from the end of the best epoch")
                     self.model.load_state_dict(self.state_dict)
 
     def on_train_end(self, logs: Optional[Dict] = None):
@@ -525,3 +653,11 @@ class EarlyStopping(Callback):
                 RuntimeWarning,
             )
         return monitor_value
+
+    def __getstate__(self):
+        d = self.__dict__
+        self_dict = {k: d[k] for k in d if k not in ["trainer", "model"]}
+        return self_dict
+
+    def __setstate__(self, state):
+        self.__dict__ = state
