@@ -16,6 +16,7 @@ from typing import (
 )
 from pathlib import PosixPath
 
+import torch
 from torch import Tensor
 from torch.nn import Module
 from torch.optim.optimizer import Optimizer
@@ -49,6 +50,10 @@ from torchvision.transforms import (
 from torch.optim.lr_scheduler import _LRScheduler
 from torch.utils.data.dataloader import DataLoader
 
+from pytorch_widedeep.models import WideDeep
+from pytorch_widedeep.models.tabnet.sparsemax import Entmax15, Sparsemax
+from pytorch_widedeep.models.transformers.layers import FullEmbeddingDropout
+
 ListRules = Collection[Callable[[str], str]]
 Tokens = Collection[Collection[str]]
 Transforms = Union[
@@ -80,3 +85,5 @@ Transforms = Union[
 ]
 LRScheduler = _LRScheduler
 ModelParams = Generator[Tensor, Tensor, Tensor]
+NormLayers = Union[torch.nn.Identity, torch.nn.LayerNorm, torch.nn.BatchNorm1d]
+DropoutLayers = Union[torch.nn.Dropout, FullEmbeddingDropout]
