@@ -1,3 +1,4 @@
+import os
 import shutil
 import string
 from copy import deepcopy
@@ -10,6 +11,7 @@ from sklearn.model_selection import train_test_split
 from pytorch_widedeep.models import (
     Wide,
     TabMlp,
+    TabNet,
     DeepText,
     WideDeep,
     DeepImage,
@@ -31,12 +33,12 @@ embed_input_tt = [(u, i) for u, i in zip(colnames[:5], [5] * 5)]
 cont_cols = [np.random.rand(32) for _ in range(5)]
 X_tab = np.vstack(embed_cols + cont_cols).transpose()
 
-#  Text Array
+# Text Array
 padded_sequences = np.random.choice(np.arange(1, 100), (32, 48))
 X_text = np.hstack((np.repeat(np.array([[0, 0]]), 32, axis=0), padded_sequences))
 vocab_size = 100
 
-#  Image Array
+# Image Array
 X_img = np.random.choice(256, (32, 224, 224, 3))
 X_img_norm = X_img / 255.0
 
@@ -78,11 +80,17 @@ tabtransformer = TabTransformer(
     embed_input=embed_input_tt,
     continuous_cols=colnames[5:],
 )
+tabnet = TabNet(
+    column_idx={k: v for v, k in enumerate(colnames)},
+    embed_input=embed_input,
+    continuous_cols=colnames[5:],
+    ghost_bn=False,
+)
 deeptext = DeepText(vocab_size=vocab_size, embed_dim=32, padding_idx=0)
 deepimage = DeepImage(pretrained=True)
 
 ###############################################################################
-#  test consistecy between optimizers and lr_schedulers format
+# test consistecy between optimizers and lr_schedulers format
 ###############################################################################
 
 
@@ -102,7 +110,7 @@ def test_optimizer_scheduler_format():
 
 
 ###############################################################################
-#  test that callbacks are properly initialised internally
+# test that callbacks are properly initialised internally
 ###############################################################################
 
 
@@ -110,11 +118,11 @@ def test_non_instantiated_callbacks():
     model = WideDeep(wide=wide, deeptabular=tabmlp)
     callbacks = [EarlyStopping]
     trainer = Trainer(model, objective="binary", callbacks=callbacks)
-    assert trainer.callbacks[1].__class__.__name__ == "EarlyStopping"
+    assert trainer.callbacks[2].__class__.__name__ == "EarlyStopping"
 
 
 ###############################################################################
-#  test that multiple metrics are properly constructed internally
+# test that multiple metrics are properly constructed internally
 ###############################################################################
 
 
@@ -129,7 +137,7 @@ def test_multiple_metrics():
 
 
 ###############################################################################
-#  test the train step with metrics runs well for a binary prediction
+# test the train step with metrics runs well for a binary prediction
 ###############################################################################
 
 
@@ -158,7 +166,7 @@ def test_basic_run_with_metrics_binary(wide, deeptabular):
 
 
 ###############################################################################
-#  test the train step with metrics runs well for a muticlass prediction
+# test the train step with metrics runs well for a muticlass prediction
 ###############################################################################
 
 
@@ -187,7 +195,7 @@ def test_basic_run_with_metrics_multiclass():
 
 
 ###############################################################################
-#  test predict method for individual components
+# test predict method for individual components
 ###############################################################################
 
 
@@ -225,7 +233,7 @@ def test_predict_with_individual_component(
 
 
 ###############################################################################
-#  test save and load
+# test save
 ###############################################################################
 
 
@@ -234,8 +242,8 @@ def test_save_and_load():
     trainer = Trainer(model, objective="binary", verbose=0)
     trainer.fit(X_wide=X_wide, X_tab=X_tab, target=target, batch_size=16)
     wide_weights = model.wide.wide_linear.weight.data
-    trainer.save_model("tests/test_model_functioning/model_dir/model.t")
-    n_model = Trainer.load_model("tests/test_model_functioning/model_dir/model.t")
+    trainer.save("tests/test_model_functioning/model_dir/")
+    n_model = torch.load("tests/test_model_functioning/model_dir/wd_model.pt")
     n_wide_weights = n_model.wide.wide_linear.weight.data
     assert torch.allclose(wide_weights, n_wide_weights)
 
@@ -259,12 +267,78 @@ def test_save_and_load_dict():
         batch_size=16,
     )
     wide_weights = model1.wide.wide_linear.weight.data
-    trainer1.save_model_state_dict("tests/test_model_functioning/model_dir/model_d.t")
+    trainer1.save(path="tests/test_model_functioning/model_dir/", save_state_dict=True)
     model2 = WideDeep(wide=wide, deeptabular=tabmlp)
     trainer2 = Trainer(model2, objective="binary", verbose=0)
-    trainer2.load_model_state_dict("tests/test_model_functioning/model_dir/model_d.t")
+    trainer2.model.load_state_dict(
+        torch.load("tests/test_model_functioning/model_dir/wd_model.pt")
+    )
     n_wide_weights = trainer2.model.wide.wide_linear.weight.data
-
+    same_weights = torch.allclose(wide_weights, n_wide_weights)
+    if os.path.isfile(
+        "tests/test_model_functioning/model_dir/history/train_eval_history.json"
+    ):
+        history_saved = True
+    else:
+        history_saved = False
     shutil.rmtree("tests/test_model_functioning/model_dir/")
+    assert same_weights and history_saved
 
-    assert torch.allclose(wide_weights, n_wide_weights)
+
+###############################################################################
+# test explain matrices and feature importance for TabNet
+###############################################################################
+
+
+def test_explain_mtx_and_feat_imp():
+    model = WideDeep(deeptabular=tabnet)
+    trainer = Trainer(model, objective="binary", verbose=0)
+    trainer.fit(
+        X_tab=X_tab,
+        target=target,
+        batch_size=16,
+    )
+
+    checks = []
+    checks.append(len(trainer.feature_importance) == len(tabnet.column_idx))
+
+    expl_mtx, step_masks = trainer.explain(X_tab[:6], save_step_masks=True)
+    checks.append(expl_mtx.shape[0] == 6)
+    checks.append(expl_mtx.shape[1] == 10)
+
+    for i in range(tabnet.n_steps):
+        checks.append(step_masks[i].shape[0] == 6)
+        checks.append(step_masks[i].shape[1] == 10)
+
+    assert all(checks)
+
+
+###############################################################################
+# test save load and predict
+###############################################################################
+
+
+def test_save_load_and_predict():
+
+    fpath = "tests/test_model_functioning/test_wd_model"
+    if not os.path.exists(fpath):
+        os.makedirs(fpath)
+
+    model = WideDeep(deeptabular=tabmlp)
+    trainer = Trainer(model, objective="binary", verbose=0)
+    trainer.fit(
+        X_tab=X_tab,
+        target=target,
+        batch_size=16,
+    )
+
+    trainer.save(path=fpath, save_state_dict=True)
+
+    model_new = WideDeep(deeptabular=tabmlp)
+    model_new.load_state_dict(torch.load("/".join([fpath, "wd_model.pt"])))
+    trainer_new = Trainer(model, objective="binary", verbose=0)
+    preds = trainer_new.predict(X_tab=X_tab, batch_size=16)
+
+    shutil.rmtree(fpath)
+
+    assert preds.shape[0] == X_tab.shape[0]
